/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Initial Developer of the Original Code is Malte Kraus.
 *
 * Contributor(s):
 *  Malte Kraus <firefox@maltekraus.de> (Original author)
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

const Ci = Components.interfaces, Cc = Components.classes, Cr = Components.results;

const SYNC_BASE_FILE = "adblockfilters_last_sync.txt";

const EXT_VERSION = "2.0"; // extension manager not available in SeaMonkey :-(

var globalObject = this;

function LOG(msg) {
  msg = "Adblock Filters:\t" + msg;
  var consoleService = Cc["@mozilla.org/consoleservice;1"]
                         .getService(Ci.nsIConsoleService);
  consoleService.logStringMessage(msg);
  //dump(msg + "\n");
  return msg;
}
function makeURI(uri, originCharset, baseURI) {
  var ioservice = Cc["@mozilla.org/network/io-service;1"]
                   .getService(Ci.nsIIOService);
  return ioservice.newURI(uri, originCharset || null, baseURI || null);
}
function loadScript(url, context) {
  if(url.indexOf(":/") == -1) url = "chrome://adblockfilters/content/" + url;
  if(!context && loadScript.caller && loadScript.caller != loadScriptOnce)
    context = loadScript.caller.__parent__;
  var ssLoader = Cc["@mozilla.org/moz/jssubscript-loader;1"]
                   .getService(Ci.mozIJSSubScriptLoader);
  ssLoader.loadSubScript(url, context || null);
}
var loadScriptOnce = (function() {
  var loaded = {};
  return function _loadScriptOnce(url, context) {
    if(!context && _loadScriptOnce.caller)
      context = _loadScriptOnce.caller.__parent__;
    if(!(url in loaded) || loaded[url] != context) {
      loaded[url] = context;
      loadScript(url, context);
    }
  };
})();
loadScriptOnce("chrome://global/content/nsUserSettings.js");


function setTimeout(callback, timeout) {
  var args = [];
  for(var i = 2; i < arguments.length; i++)
    args.push(arguments[i]);
  var obj = { notify: function() { return callback.apply(timer, args); }};
  obj.wrappedJSObject = obj;
  var timer = Cc["@mozilla.org/timer;1"].createInstance(Ci.nsITimer);
  timer.initWithCallback(obj, timeout, Ci.nsITimer.TYPE_ONE_SHOT);
  return {t:timer};
}
function clearTimeout(timeout) {
  var timer = timeout.t;
  timer.cancel();
  // prevent cycle references:
  timer.callback.wrappedJSObject.notify = null;
  delete timeout.t;
}

// add prototype to Constructor's prototype chain, both may already have any number of prototypes
function addToPrototypeChain(Constructor, prototype) {
  var proto = Constructor.prototype;

  // walk the prototype chain to its end:
  while(proto.__proto__ != Object.prototype) {
    proto = proto.__proto__;
  }
  // add the new prototype at the end of the current chain:
  proto.__proto__ = prototype;
}

var adblockFilterServiceInstance;
function AdblockFilterService() {
  if(adblockFilterServiceInstance)
    return adblockFilterServiceInstance;
  adblockFilterServiceInstance = this.wrappedJSObject = this;
  this.prefs = nsPreferences;
  this._startMigration();
  this._registerPrefObserver();
  this._startSyncTimer();
}
AdblockFilterService.prototype = {
  classID: Components.ID("{76c938ef-4597-2f95-4f1a-fad72f7ecb6c}"),
  contractID: "@adblockfilters.mozdev.org/service;1",
  classDescription: "Adblock Plus Filter Uploader service",

  wrappedJSObject: null,
  prefs: null,

  _startMigration: function() {
    const PREFIX = "extensions.adblockfilters.";
    var version = this.prefs.copyUnicharPref(PREFIX + "lastVersion");
    if(version != EXT_VERSION) {
      if(version == "0") {
        // we didn't set this pref until version 2.0
        // use a set host to identify prior versions
        if(this.prefs.copyUnicharPref(PREFIX + "ftp.host"))
          version = "1.*";
      }
      if(version == "1.*") {
        // synchronization wasn't available in 1.x, so don't break these users
        this.prefs.setBoolPref(PREFIX + "sync.enabled", false);
      }
      this.prefs.setUnicharPref(PREFIX + "lastVersion", EXT_VERSION);
    }
  },

  _registerCloseObserver: function() {
    Cc["@mozilla.org/observer-service;1"]
      .getService(Ci.nsIObserverService)
      .addObserver(this, "domwindowclosed", false);
  },
  _unregisterCloseObserver: function() {
    Cc["@mozilla.org/observer-service;1"]
      .getService(Ci.nsIObserverService)
      .removeObserver(this, "domwindowclosed");
  },
  _registerPrefObserver: function() {
    const PREFIX = "extensions.adblockfilters.";
    if(this.prefs.getBoolPref(PREFIX + "uploadOnExit"))
      this._registerCloseObserver();
    this.prefs.mPrefService.QueryInterface(Ci.nsIPrefService)
        .getBranch(PREFIX).QueryInterface(Ci.nsIPrefBranch2)
        .addObserver(PREFIX, this, false);
  },

  observe: function (aSubject, aTopic, aData) {
    if(aTopic == "nsPref:changed") {
      switch(aData) {
        case "uploadOnExit":
          if(this.prefs.getBoolPref("uploadOnExit"))
            this._registerCloseObserver();
          else
            this._unregisterCloseObserver();
          break;
        case "sync.enabled":
        case "sync.interval":
        case "sync.lastSync":
          this._startSyncTimer();
          break;
      }
      if(aData.indexOf("blacklist.") == 0) {
        this._blacklist = null; // regenerate blacklist lazily
      }
    } else if(aTopic == "domwindowclosed") {
      this._doLastUpload(aSubject);
    }
  },

  _doLastUpload: function(aWindow) {
    let win = Cc["@mozilla.org/appshell/window-mediator;1"]
                .getService(Ci.nsIWindowMediator).getMostRecentWindow("");
    if(win) // there's at least one remaining window open
      return;

    this._unregisterCloseObserver();

    this.doUpload(null, true);
  },

  /* utility functions */
  openDialog: function(window, windowType, url, features, args) {
    if(windowType) {
      let wm = Cc["@mozilla.org/appshell/window-mediator;1"].getService(Ci.nsIWindowMediator);
      let win = wm.getMostRecentWindow(windowType);
      if(win) {
        win.focus();
        return win;
      }
    }

    let argsArr = null;
    if(args) {
      let Variant = Components.Constructor("@mozilla.org/variant;1",
                                          "nsIWritableVariant", "setFromVariant");
      argsArr = Cc["@mozilla.org/array;1"].createInstance(Ci.nsIMutableArray);
      args.forEach(function(argument) {
        argsArr.appendElement(new Variant(argument), false);
      });
    }

    let ww = Cc["@mozilla.org/embedcomp/window-watcher;1"]
                   .getService(Ci.nsIWindowWatcher);
    return ww.openWindow(window || null, url, url, features, argsArr);
  },
  doUpload: function(aWindow, aSilent) {
    if(Cc["@mozilla.org/appshell/window-mediator;1"].getService(Ci.nsIWindowMediator)
         .getMostRecentWindow("adblockfilters:uploader"))
      return; // don't allow multiple instances

    let url = "", args = null;
    if(this.syncEnabled && // sync not initialized
       nsPreferences.getIntPref("extensions.adblockfilters.sync.version") == 0) {
      if(aSilent)
        return;
      url = "first-sync.xul";
    } else {
      url = "ftp-upload.xul";
      let up = new SyncJob(this.getData());
      if(!up.sync && !up.dataChanged) // we don't have to do anything
        return;
      args = [up, aSilent];
    }
    this.openDialog(aWindow, "", "chrome://adblockfilters/content/" + url, "chrome", args);
  },
  initSynchronization: function() {
    let ww = Cc["@mozilla.org/embedcomp/window-watcher;1"]
                   .getService(Ci.nsIWindowWatcher);
    const URL = "chrome://adblockfilters/content/first-sync.xul";
    return ww.openWindow(null, URL, URL, "chrome", null);
  },
  openURL: function openURL(URL) {
    var appInfo = Cc["@mozilla.org/xre/app-info;1"]
                    .createInstance(Ci.nsIXULAppInfo);
    switch(appInfo.ID) {
      case "{ec8030f7-c20a-464f-9b0e-13a3a9e97384}": // Firefox
      case "{92650c4d-4b8e-4d2a-b7eb-24ecf4f6b63a}": // SeaMonkey
      case "{a463f10c-3994-11da-9945-000d60ca027b}": // Flock
      case "{3db10fab-e461-4c80-8b97-957ad5f8ea47}": // Netscape
        let win = Cc["@mozilla.org/appshell/window-mediator;1"]
                    .getService(Ci.nsIWindowMediator)
                    .getMostRecentWindow("navigator:browser");

        let TAB_PREF = "browser.link.open_external";
        let tabsEnabled = nsPreferences.getIntPref(TAB_PREF) == 3;

        if(tabsEnabled && win) {
          win.focus();
          let browser = win.getBrowser();
          if(browser.selectedBrowser.currentURI.spec != "about:blank")
            browser.selectedTab = browser.addTab(URL);
          else
            browser.loadURI(makeURI(URL));
        } else {
          let windowFeatures = "chrome=0,resizable,modal=0,dialog=0,close,"
                             + "minimizable,titlebar,menubar,toolbar,location,"
                             + "directories,personalbar,status";
          win.open(URL, "_blank", windowFeatures);
        }
        break;
      case "{3550f703-e582-4d05-9a08-453d09bdfdc6}": // Thunderbird
      case "{718e30fb-e89b-41dd-9da7-e25a45638b28}": // Sunbird
      case "{136c295a-4a5a-41cf-bf24-5cee526720d5}": // NVU
      case "songbird@songbirdnest.com":              // Songbird
      default: // open in OS default
        Cc["@mozilla.org/uriloader/external-protocol-service;1"]
          .getService(Ci.nsIExternalProtocolService).loadUrl(makeURI(URL));
        break;
    }
  },
  get abpVersion() {
    if("@mozilla.org/adblockplus;1" in Cc)
      return Cc["@mozilla.org/adblockplus;1"]
               .createInstance().wrappedJSObject.getInstalledVersion();
    else
      return "0";
  },

  writeFile: function(fileName, data, raw) {
    let file = Cc["@mozilla.org/file/directory_service;1"]
                 .getService(Ci.nsIProperties).get("ProfD", Ci.nsIFile);
    file.append(fileName);
    let output = Cc["@mozilla.org/network/file-output-stream;1"]
                   .createInstance(Ci.nsIFileOutputStream);
    output.init(file, 0x02 | 0x08 | 0x20, 0600, 0);

    if(raw) {
      let stream = Cc["@mozilla.org/binaryoutputstream;1"]
                  .createInstance(Ci.nsIBinaryOutputStream);
      stream.setOutputStream(output);
      stream.writeBytes(data, data.length);
      stream.close();
    } else {
      let converter = Cc["@mozilla.org/intl/converter-output-stream;1"]
                  .createInstance(Ci.nsIConverterOutputStream);
      converter.init(output, "UTF-8", 0, 0);
      converter.writeString(data);
      converter.close();
    }
  },
  readFile: function(fileName) {
    let file = Cc["@mozilla.org/file/directory_service;1"]
                 .getService(Ci.nsIProperties).get("ProfD", Ci.nsIFile);
    file.append(fileName);
    if(!file.exists() || !file.isReadable())
      return "";

    let fstream = Cc["@mozilla.org/network/file-input-stream;1"]
                    .createInstance(Ci.nsIFileInputStream);
    fstream.init(file, 0x01, 0600, Ci.nsIFileInputStream.CLOSE_ON_EOF);
    let is = Cc["@mozilla.org/intl/converter-input-stream;1"]
               .createInstance(Ci.nsIConverterInputStream);
    is.init(fstream, "UTF-8", 0, 0);

    let str = "";
    while(fstream.available())
      str += let(tmp = {}) (is.readString(fstream.available(), tmp), tmp.value);
    return str;
  },

  // nsISupports
  QueryInterface: function(aIID) {
    if(aIID.equals(Ci.nsISupports) || aIID.equals(Ci.nsIClassInfo) ||
       aIID.equals(Ci.nsIObserver))
      return this;
    throw Cr.NS_ERROR_NO_INTERFACE;
  },

  // nsIClassInfo
  getHelperForLanguage: function() {
    return null;
  },
  getInterfaces: function getInterfaces(aCount) {
    var array = [Ci.nsIClassInfo, Ci.nsIObserver];
    aCount.value = array.length;
    return array;
  },
  classIDNoAlloc: Ci.nsISupports,
  flags: Ci.nsIClassInfo.SINGLETON,
  implementationLanguage: Ci.nsIProgrammingLanguage.JAVASCRIPT
};


loadScript("collect-filters.js"); // collect abp filters
addToPrototypeChain(AdblockFilterService, new FilterCollector());

loadScript("timer-sync.js"); // synchronize every X hours
addToPrototypeChain(AdblockFilterService, new SyncTimer());

loadScript("sync-job.js"); // synchronization "job"


// constructors for objects we want to XPCOMify
// their prototype must have the properties classID, contractID, classDescription
var XPCOM_Constructors = [AdblockFilterService, SyncJob];
/*
 * Common registration code.
 *
 */

  var module = {
    registerSelf: function (aComponentManager, aFileSpec, aLocation, aType) {
      aComponentManager.QueryInterface(Ci.nsIComponentRegistrar);
      for each(let obj in this._objects) {
        aComponentManager.registerFactoryLocation(obj.classID, obj.classDescription,
                                     obj.contractID, aFileSpec, aLocation, aType);
      }
    },
    unregisterSelf: function(aCompMgr, aFileSpec, aLocation) {
      aComponentManager.QueryInterface(Ci.nsIComponentRegistrar);
      for each(let obj in this._objects) {
        aComponentManager.unregisterFactoryLocation(obj.classID, aFileSpec);
      }
    },
    getClassObject: function(aComponentManager, aCID, aIID) {
      if (!aIID.equals(Ci.nsIFactory)) throw CR.NS_ERROR_NOT_IMPLEMENTED;

      var key = aCID.number;
      if (key in this._objects && aCID.equals(this._objects[key].classID))
        return this._objects[key].factory;

      throw CR.NS_ERROR_NO_INTERFACE;
    },
    canUnload: function(aComponentManager) { return true; },
    _objects: {} //FactoryHolder
  };

function NSGetModule(compMgr, fileSpec) {
  function FactoryHolder(aObj) {
    this.factory = {
      createInstance: function(aOuter, aIID) {
        if(aOuter)
          throw Cr.NS_ERROR_NO_AGGREGATION;
        return (new aObj()).QueryInterface(aIID);
      }
    };
    this.__proto__ = aObj.prototype; // make classID, contractID, classDescription available for our users
  }

  XPCOM_Constructors.forEach(function(constructor) {
    module._objects[constructor.prototype.classID.number] = new FactoryHolder(constructor);
  });
  return module;
}